// -----------------------------------------------------------------------
//   
//   Copyright 2008 H. Peter Anvin - All Rights Reserved
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
//   Boston MA 02110-1301, USA; either version 2 of the License, or
//   (at your option) any later version; incorporated herein by reference.
//
// -----------------------------------------------------------------------

//
// Sound DAC I2C controller
//
// The I2S DAC on the DE1 board need to be initialized via I2C.
//

module sound_i2c (
		  input rst_n,	 	// Reset
		  input clk,	 	// System clock
		  input clk_en,		// Strobe (<= 1.6 MHz)

		  inout i2c_scl,	// I2C SCL
		  inout i2c_sda		// I2C SDA
		  );

   parameter 		i2c_addr = 7'h1A; // DAC I2C address

   // Only need to set registers 0-9
   parameter 		i2c_rom_stop = 4'd10;
      
   reg   [3:0] 		i2c_rom_a;
   wire  [8:0] 		i2c_rom_q;
   reg  [26:0] 		i2c_bit_shr;
   reg   [3:0] 		i2c_scl_out;
   reg   [3:0] 		i2c_sda_out;
   reg   [4:0] 		i2c_bit_ctr;
   reg   [1:0] 		i2c_qtr;

   reg 			i2c_active = 1'b0;

   wire 		i2c_stop = i2c_rom_a == i2c_rom_stop;
   
   // Output drivers: all outputs are open drain
   assign 		i2c_scl = i2c_scl_out[3] ? 1'bz : 1'b0;
   assign 		i2c_sda = i2c_sda_out[3] ? 1'bz : 1'b0;

   sound_i2c_rom sound_i2c_rom (
				.clk (clk),
				.a (i2c_rom_a),
				.q (i2c_rom_q)
				);
 
   // We send packets of 3x9 = 27 bits, plus S and P conditions.

   always @(posedge clk)
     if ( ~rst_n )
       begin
	  // Synchronous reset
	  i2c_rom_a   <= 4'd0;
	  i2c_bit_shr <= 27'hxxxx;
	  i2c_bit_ctr <= 5'd0;
	  i2c_qtr     <= 2'b00;
	  i2c_active  <= 1'b1;
	  i2c_scl_out <= 4'b1111;
	  i2c_sda_out <= 4'b1111;
       end
     else if ( clk_en & i2c_active & (~i2c_scl_out[3] | i2c_scl) )
       begin
	  i2c_scl_out <= { i2c_scl_out[2:0], 1'b1 };
	  i2c_sda_out <= { i2c_sda_out[2:0], 1'b1 };
	  i2c_qtr     <= i2c_qtr + 1;

	  if ( i2c_qtr == 2'b00 )
	    begin
	       i2c_bit_ctr <= i2c_bit_ctr + 1;
	       i2c_bit_shr <= { i2c_bit_shr[25:0], 1'bx };

	       casex ( i2c_bit_ctr )
		 5'b00000:
		   begin
		      // Start condition, prepare for data transfer
		      i2c_scl_out <= 4'b1110;
		      i2c_sda_out <= 4'b1000;
		      i2c_bit_shr <= { i2c_addr, 1'b0, 1'b1,
				       3'b000, i2c_rom_a, i2c_rom_q[8], 1'b1,
				       i2c_rom_q[7:0], 1'b1 };
		      i2c_rom_a   <= i2c_rom_a + 1;
		      i2c_active  <= ~i2c_stop;
		   end
		 5'b111xx:
		   begin
		      // Stop condition
		      i2c_scl_out <= 4'b0111;
		      i2c_sda_out <= 4'b0001;
		      i2c_bit_ctr <= 5'd0;
		   end
		 default:
		   begin
		      // Send data bit
		      i2c_scl_out <= 4'b0110;
		      i2c_sda_out <= {4{i2c_bit_shr[26]}};
		   end
	       endcase // casex( i2c_bit_ctr )
	    end // if ( i2c_qtr == 2'b00 )
       end // if ( clk_en & i2c_active & (~i2c_scl_out[3] | i2c_scl) )

endmodule // sound_i2c

// It should be possible to infer a ROM from this construct, but
// since the number of inputs are so small, the synthesizer probably
// uses LUTs as ROMs instead.
module sound_i2c_rom (
		      input        clk,
		      input  [3:0] a,
		      output [8:0] q
	              );

   reg  [3:0]    a_reg;
   reg  [8:0]    q_reg;

   assign 	 q = q_reg;
   
   always @(posedge clk)
     a_reg <= a;

   always @(posedge clk)
     case (a_reg)
       4'h0:    q_reg <= 9'h01a;
       4'h1:    q_reg <= 9'h01a;
       4'h2:    q_reg <= 9'h07b;
       4'h3:    q_reg <= 9'h07b;
       4'h4:    q_reg <= 9'h0f8;
       4'h5:    q_reg <= 9'h006;
       4'h6:    q_reg <= 9'h000;
       4'h7:    q_reg <= 9'h001;
       4'h8:    q_reg <= 9'h002;
       4'h9:    q_reg <= 9'h001;
       default: q_reg <= 9'hxxx;
     endcase
   
endmodule // sound_i2c_rom
